/*!
 * \file particle.cc
 * \brief Implements ParticleFilter and AdaptiveParticleFilter classes.
 * \author Boyoon Jung (boyoon@robotics.usc.edu)
 */
#include "particle.h"
using namespace bj;

#include <cmath>

// header files for debugging
#include <iostream>
using std::cerr;
using std::endl;


//-----------------------------------------------------------------------------
// for ParticleFilter class
//-----------------------------------------------------------------------------

// constructor
ParticleFilter::ParticleFilter(int num_particles, int dimension,
			       gsl_vector* min, gsl_vector* max)
{
    // initialize a random number generator
    rng = gsl_rng_alloc(gsl_rng_taus);
    timeval tv;
    gettimeofday(&tv, NULL);		// initialize it using the system timer
    gsl_rng_set(rng, tv.tv_usec);

    // store the parameters
    size = num_particles;
    p_size = dimension;

    ss_min = gsl_vector_alloc(p_size);
    if (min) gsl_vector_memcpy(ss_min, min);
    else gsl_vector_set_zero(ss_min);

    ss_max = gsl_vector_alloc(p_size);
    if (max) gsl_vector_memcpy(ss_max, max);
    else gsl_vector_set_all(ss_max, 1.0);

    // allocate particles
    particle = new gsl_vector*[size];
    importance = new double[size];
    for (int i=0; i<size; i++) particle[i] = gsl_vector_alloc(p_size);

    // allocate internal buffer
    cdf = new double[size+1];
    p_tmp = new gsl_vector*[size];
    for (int i=0; i<size; i++) p_tmp[i] = gsl_vector_alloc(p_size);

    // initialize the particle filter
    initialize();
}


// destructor
ParticleFilter::~ParticleFilter(void)
{
    // de-allocate the internal buffer
    delete[] cdf;
    for (int i=0; i<size; i++) gsl_vector_free(p_tmp[i]);
    delete[] p_tmp;
    
    // de-allocate particles
    for (int i=0; i<size; i++) gsl_vector_free(particle[i]);
    delete[] particle;
    delete[] importance;

    // de-allocate limit objects
    gsl_vector_free(ss_min);
    gsl_vector_free(ss_max);

    // de-allocate the random number generator
    gsl_rng_free(rng);
}


// perform samplling-importance-resampling
int ParticleFilter::sir(void)
{
    // perform an uniform sampling along y-axis
    double rn = gsl_rng_uniform(rng);

    // search the corresponding particle on x-axis (binary search)
    int start=0;
    int end = size;
    while ((start+1) < end)
    {
	int center = (start + end) / 2;
	if (rn < cdf[center])
	    end = center;
	else
	    start = center;
    }

    // done
    return start;
}


// initialize a particle filter
void ParticleFilter::initialize(void)
{
    for (int i=0; i<size; i++)
    {
	// generate totally random particles
	set_particle_random(particle[i]);

	// set all importances uniformly
	importance[i] = 1.0 / size;

	// calculate CDF
	cdf[i] = i * 1.0 / size;
    }
    cdf[size] = 1.0;
}


// update particles
bool ParticleFilter::update(void)
{
    int p_idx;

    // generate a new particle set
    double i_sum = 0;
    for (int i=0; i<size; i++)
    {
	// draw a particle according to importance factors
	p_idx = sir();

	// propagate the particle through an action model
	propagate(particle[p_idx], p_tmp[i]);

	// evaluate the new particle based on a measurement model
	i_sum += importance[i] = evaluate(p_tmp[i]);
    }

    // check if all particles are invalid
    if (i_sum == 0.0) {
	cerr << "[Warning] none of the particles is valid. "
	     << "the particle filter is reset." << endl;
	initialize();
	return false;
    }

    // normalize importance factors and calculate CDF
    cdf[0] = 0;
    for (int i=0; i<size; i++)
    {
	// normalize importance factors
	importance[i] /= i_sum;

	// calculate CDF
	cdf[i+1] = cdf[i] + importance[i];
    }

    // swap data buffers
    gsl_vector** ps = particle;
    particle = p_tmp;
    p_tmp = ps;

    // done
    return true;
}



//-----------------------------------------------------------------------------
// for AdaptiveParticleFilter class
//-----------------------------------------------------------------------------

// constructor
AdaptiveParticleFilter::AdaptiveParticleFilter(int min_particles, int max_particles,
					       int dimension,
					       gsl_vector* min, gsl_vector* max,
					       gsl_vector* bsize)
	: ParticleFilter(max_particles, dimension, min, max)
{
    // store parameters
    if (min_particles > 1)
	min_size = min_particles;
    else {
	cerr << "[Warning] more than 1 particles are required at least." << endl;
	min_size = 2;
    }

    if (max_particles >= min_particles)
	max_size = max_particles;
    else {
	cerr << "[Warning] max_particles should be bigger than min_particles." << endl;
	max_size = min_size;
    }

    if (bsize) {
	bin_size = gsl_vector_alloc(bsize->size);
	gsl_vector_memcpy(bin_size, bsize);
    }
    else {
	bin_size = gsl_vector_alloc(dimension);
	gsl_vector_set_all(bin_size, 1.0);		// default bin size
    }

    // create a kd-tree
    kd_tree = new KDTree(bin_size);

    // initialize parameters
    size = max_size;		// start with the maximum number of particles
    err_bound = 0.01;
    conf_quantile = 2.0537489;	// 98% confidence
}


// destructor
AdaptiveParticleFilter::~AdaptiveParticleFilter(void)
{
    // de-allocate objects
    gsl_vector_free(bin_size);

    // before the destructor of ParticleFilter class is invoked
    size = max_size;
}


// update particles
bool AdaptiveParticleFilter::update(void)
{
    int p_idx;

    // initialize a kd-tree
    kd_tree->reset();

    // generate a new particle set
    int new_size = 0;
    double i_sum = 0;
    do
    {
	// draw a particle according to importance factors
	p_idx = sir();

	// propagate the particle through an action model
	propagate(particle[p_idx], p_tmp[new_size]);

	// evaluate the new particle based on a measurement model
	i_sum += importance[new_size] = evaluate(p_tmp[new_size]);

	// insert the new particle into a KD-tree
	kd_tree->insert(p_tmp[new_size]);

	// increase the counter
	new_size++;

    } while (new_size < min_size ||
	     (new_size < max_size && new_size < proper_size(kd_tree->size())));

    // update the number of particles
    size = new_size;

    // check if all particles are invalid
    if (i_sum == 0.0) {
	cerr << "[Warning] none of the particles is valid. "
	     << "the particle filter is reset." << endl;
	initialize();
	return false;
    }

    // calculate importance factors and CDF
    cdf[0] = 0;
    for (int i=0; i<size; i++)
    {
	// normalize importance factors
	importance[i] /= i_sum;

	// calculate CDF
	cdf[i+1] = cdf[i] + importance[i];
    }

    // swap data buffers
    gsl_vector** ps = particle;
    particle = p_tmp;
    p_tmp = ps;

    // done
    return true;
}

